#include "QrssPiG.h"

#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <string>
#include <syslog.h>

#include <yaml-cpp/yaml.h>

#include "QGPlugin.h"

using std::placeholders::_1;
using std::placeholders::_2;
using std::placeholders::_3;
using std::placeholders::_4;
using std::placeholders::_5;

void QrssPiG::listModules() {
	std::cout << "Plugin path:" << std::endl;
	std::cout << "\t" << QGPlugin<QGInputDevice>::plugin_path() << std::endl;
	std::cout << std::endl;

	std::cout << "Available input plugins:" << std::endl;
	for (const auto&m: QGInputDevice::listModules()) std::cout << "\t" << std::left << std::setw(12) << std::string(m.first + ":") << m.second << std::endl;
	std::cout << std::endl;

	std::cout << "Available upload plugins:" << std::endl;
	for (const auto&m: QGUploader::listModules()) std::cout << "\t" << m << std::endl;
	std::cout << std::endl;
}

void QrssPiG::listDevices() {
	auto devices = QGInputDevice::listDevices();

	std::cout << "Available input devices:" << std::endl;
	for (const auto& m: devices) {
		if (m.second.size()) for (const auto &d: m.second) std::cout << "\t" << std::left << std::setw(12) << std::string(m.first + ":") << d << std::endl;
		else std::cout << "\t" << std::left << std::setw(12) << std::string(m.first + ":") << "No device found" << std::endl;
	}
	std::cout << std::endl;
}

void QrssPiG::dummyOutput(const std::string &configFile) {
	YAML::Node config = YAML::LoadFile(configFile);

	// Create input device
	if (config["input"] && config["input"].Type() != YAML::NodeType::Map)
		throw std::runtime_error("YAML: input must be a map");

	config["input"]["nodevice"] = true;
	std::shared_ptr<QGInputDevice> device = QGInputDevice::CreateInputDevice(config["input"]);

	// Create processor
	if (config["processing"] && config["processing"].Type() != YAML::NodeType::Map)
		throw std::runtime_error("YAML: processing must be a map");

	config["processing"]["noprocessor"] = true;
	std::shared_ptr<QGProcessor> processor(new QGProcessor(config["processing"], *(device.get())));

	// Connect input to processor
	device->setCb(std::bind(&QGProcessor::addIQ, processor, _1), processor->chunkSize());

	// Create all outputs
	std::vector<std::shared_ptr<QGOutput>> outputs;

	if (config["output"]) {
		if (config["output"].Type() == YAML::NodeType::Map)
			outputs.push_back(QGOutput::CreateOutput(config["output"], *(processor.get())));
		else if (config["output"].Type() == YAML::NodeType::Sequence)
			for (unsigned int i = 0; i < config["output"].size(); i++)
				outputs.push_back(QGOutput::CreateOutput(config["output"][i], *(processor.get()), i));
		else
			throw std::runtime_error("YAML: output must be a map or a sequence");
	}

	// Connect processor to every output
	for (auto&& output: outputs) processor->addCb(std::bind(&QGOutput::addLine, output, _1));

	// Create all uploader
	std::vector<std::shared_ptr<QGUploader>> uploaders;

	if (config["upload"]) {
		if (config["upload"].IsMap())
			uploaders.push_back(QGUploader::CreateUploader(config["upload"]));
		else if (config["upload"].IsSequence())
			for (YAML::const_iterator it = config["upload"].begin(); it != config["upload"].end(); it++)
				uploaders.push_back(QGUploader::CreateUploader(*it));
		else throw std::runtime_error("YAML: upload must be a map or a sequence");
	}

	// Connect every output to every uploader
	for (auto&& uploader: uploaders)
		for (auto&& output: outputs)
			output->addCb(std::bind(&QGUploader::push, uploader, _1, _2, _3, _4, _5));
}

QrssPiG::QrssPiG(const std::string &configFile) {
	YAML::Node config = YAML::LoadFile(configFile);

	syslog(LOG_INFO, "Configuring");

std::cout << "Configuring input" << std::endl;
	// Create input device
	if (config["input"] && config["input"].Type() != YAML::NodeType::Map)
		throw std::runtime_error("YAML: input must be a map");
	_inputDevice = QGInputDevice::CreateInputDevice(config["input"]);

std::cout << "Configuring processing" << std::endl;
	// Create processor
	if (config["processing"] && config["processing"].Type() != YAML::NodeType::Map)
		throw std::runtime_error("YAML: processing must be a map");
	_processor.reset(new QGProcessor(config["processing"], *(_inputDevice.get())));

	// Connect input to processor
	_inputDevice->setCb(std::bind(&QGProcessor::addIQ, _processor, _1), _processor->chunkSize());

std::cout << "Configuring output" << std::endl;
	// Create all outputs
	if (config["output"]) {
		if (config["output"].Type() == YAML::NodeType::Map)
			_outputs.push_back(QGOutput::CreateOutput(config["output"], *(_processor.get())));
		else if (config["output"].Type() == YAML::NodeType::Sequence)
			for (unsigned int i = 0; i < config["output"].size(); i++)
				_outputs.push_back(QGOutput::CreateOutput(config["output"][i], *(_processor.get()), i));
		else
			throw std::runtime_error("YAML: output must be a map or a sequence");
	}

	// Connect processor to every output
	for (auto&& output: _outputs) _processor->addCb(std::bind(&QGOutput::addLine, output, _1));

std::cout << "Configuring upload" << std::endl;
	// Create all uploader
	if (config["upload"]) {
		if (config["upload"].IsMap())
			_uploaders.push_back(QGUploader::CreateUploader(config["upload"]));
		else if (config["upload"].IsSequence())
			for (YAML::const_iterator it = config["upload"].begin(); it != config["upload"].end(); it++)
				_uploaders.push_back(QGUploader::CreateUploader(*it));
		else throw std::runtime_error("YAML: upload must be a map or a sequence");
	}

	// Connect every output to every uploader
	for (auto&& uploader: _uploaders)
		for (auto&& output: _outputs)
			output->addCb(std::bind(&QGUploader::push, uploader, _1, _2, _3, _4, _5));

std::cout << "Configured" << std::endl;
}

QrssPiG::~QrssPiG() {
}

void QrssPiG::run() {
	syslog(LOG_INFO, "Started");
	std::cout << "Started" << std::endl;
	_inputDevice->run();
}

void QrssPiG::stop() {
	if (_lastCtrlC.count()) {
		auto now = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch());

		if ((now - _lastCtrlC).count() < 1000) {
			syslog(LOG_INFO, "Forced stop");
			std::cout << "Forced stop" << std::endl;
			exit(0);
		} else {
			syslog(LOG_INFO, "Already stopping... Press Ctrl-C a second time in less than 1 second for forced stop");
			std::cout << "Already stopping... Press Ctrl-C a second time in less than 1 second for forced stop" << std::endl;
		}

		_lastCtrlC = now;
	} else {
		_lastCtrlC = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now().time_since_epoch());
		syslog(LOG_INFO, "Stopping... (press Ctrl-C a second time in less than 1 second for)");
		std::cout << "Stopping" << std::endl;
		_inputDevice->stop();
		syslog(LOG_INFO, "Stopped");
		std::cout << "Stopped" << std::endl;
	}
}
